# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Tests changelog output."""

# Standard library imports
import os
import sys
import tempfile

# Third party imports
from mock import MagicMock, patch
import pytest

# Local imports
from loghub.cli.main import create_changelog, parse_arguments
from loghub.external.github import JsonObject

REPO = 'spyder-ide/loghub'
TEST_TOKEN = os.environ.get('TEST_TOKEN', '').replace('x', '')
TEST_USER = os.environ.get('TEST_USER', '').replace('x', '')
TEST_PASS = os.environ.get('TEST_CODE', '').replace('x', '')
TEST_MILESTONE = 'test-milestone'
TEST_MILESTONE_REAL = 'v0.2'
TEST_TAG = 'v0.1.2'
NOT_ON_CI = os.environ.get('CIRCLECI') != 'true'


# --- Tests
# -----------------------------------------------------------------------------
@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_changelog_loghub_real_milestone():
    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE_REAL,
        branch='master',
        output_format='changelog')
    expected = '''## Version 0.2 (2017-02-01)

### Issues Closed

* [Issue 64](https://github.com/spyder-ide/loghub/issues/64) - Create release for v0.2
* [Issue 57](https://github.com/spyder-ide/loghub/issues/57) - Allow Enable/Disable of PR output ([PR 65](https://github.com/spyder-ide/loghub/pull/65))
* [Issue 56](https://github.com/spyder-ide/loghub/issues/56) - Fix output for grouped labels if no items on that group ([PR 58](https://github.com/spyder-ide/loghub/pull/58))
* [Issue 55](https://github.com/spyder-ide/loghub/issues/55) - Fix date on generation based on milestone ([PR 58](https://github.com/spyder-ide/loghub/pull/58))
* [Issue 53](https://github.com/spyder-ide/loghub/issues/53) - Normalize CI files using requirements.txt ([PR 60](https://github.com/spyder-ide/loghub/pull/60))
* [Issue 52](https://github.com/spyder-ide/loghub/issues/52) - Update README to use full commands ([PR 58](https://github.com/spyder-ide/loghub/pull/58))
* [Issue 47](https://github.com/spyder-ide/loghub/issues/47) - Update Appveyor Badge for new username spyder-ide ([PR 48](https://github.com/spyder-ide/loghub/pull/48))
* [Issue 46](https://github.com/spyder-ide/loghub/issues/46) - Create labels automagically on github based on template ([PR 59](https://github.com/spyder-ide/loghub/pull/59))
* [Issue 42](https://github.com/spyder-ide/loghub/issues/42) - Generate all changelog based on milestones
* [Issue 41](https://github.com/spyder-ide/loghub/issues/41) - Add default template for release with issue label grouping ([PR 43](https://github.com/spyder-ide/loghub/pull/43))
* [Issue 40](https://github.com/spyder-ide/loghub/issues/40) - Provide available milestones to user when provided not found ([PR 44](https://github.com/spyder-ide/loghub/pull/44))
* [Issue 39](https://github.com/spyder-ide/loghub/issues/39) - Add checkignore to quantified code
* [Issue 30](https://github.com/spyder-ide/loghub/issues/30) - Add label groupings ([PR 34](https://github.com/spyder-ide/loghub/pull/34))
* [Issue 28](https://github.com/spyder-ide/loghub/issues/28) - Generate all changelog based on tags ([PR 61](https://github.com/spyder-ide/loghub/pull/61))
* [Issue 27](https://github.com/spyder-ide/loghub/issues/27) - Generate all changelog based on milestones ([PR 61](https://github.com/spyder-ide/loghub/pull/61))
* [Issue 23](https://github.com/spyder-ide/loghub/issues/23) - Update Readme with new CLI parameters ([PR 36](https://github.com/spyder-ide/loghub/pull/36))
* [Issue 21](https://github.com/spyder-ide/loghub/issues/21) - Add branch parameter to select PRs to display based on the merging branch ([PR 32](https://github.com/spyder-ide/loghub/pull/32))
* [Issue 16](https://github.com/spyder-ide/loghub/issues/16) - Better authorization handling  ([PR 44](https://github.com/spyder-ide/loghub/pull/44))
* [Issue 15](https://github.com/spyder-ide/loghub/issues/15) - Since tag not working properly ([PR 13](https://github.com/spyder-ide/loghub/pull/13))
* [Issue 11](https://github.com/spyder-ide/loghub/issues/11) - Support hidden password ([PR 14](https://github.com/spyder-ide/loghub/pull/14))
* [Issue 10](https://github.com/spyder-ide/loghub/issues/10) - Support custom line formats [templates] ([PR 20](https://github.com/spyder-ide/loghub/pull/20))
* [Issue 8](https://github.com/spyder-ide/loghub/issues/8) - Support "all issues"  ([PR 19](https://github.com/spyder-ide/loghub/pull/19))
* [Issue 4](https://github.com/spyder-ide/loghub/issues/4) - Expose labels for additional filtering ([PR 17](https://github.com/spyder-ide/loghub/pull/17))
* [Issue 3](https://github.com/spyder-ide/loghub/issues/3) - Add access token interface ([PR 18](https://github.com/spyder-ide/loghub/pull/18))
* [Issue 2](https://github.com/spyder-ide/loghub/issues/2) - Add some basic tests ([PR 29](https://github.com/spyder-ide/loghub/pull/29))

In this release 25 issues were closed.

### Pull Requests Merged

* [PR 65](https://github.com/spyder-ide/loghub/pull/65) - PR: Add support for no-prs on output ([57](https://github.com/spyder-ide/loghub/issues/57))
* [PR 61](https://github.com/spyder-ide/loghub/pull/61) - PR: Add batch mode for all tags for all milestones ([28](https://github.com/spyder-ide/loghub/issues/28), [27](https://github.com/spyder-ide/loghub/issues/27))
* [PR 60](https://github.com/spyder-ide/loghub/pull/60) - PR: Add requirements file and update CI process ([53](https://github.com/spyder-ide/loghub/issues/53))
* [PR 59](https://github.com/spyder-ide/loghub/pull/59) - PR: Add a label creator utility ([46](https://github.com/spyder-ide/loghub/issues/46))
* [PR 58](https://github.com/spyder-ide/loghub/pull/58) - PR: Refactor code and simplify group handling ([56](https://github.com/spyder-ide/loghub/issues/56), [55](https://github.com/spyder-ide/loghub/issues/55), [52](https://github.com/spyder-ide/loghub/issues/52))
* [PR 54](https://github.com/spyder-ide/loghub/pull/54) - PR: Remove versioneer
* [PR 51](https://github.com/spyder-ide/loghub/pull/51) - PR: Update ignore file to ignore versioneer files
* [PR 50](https://github.com/spyder-ide/loghub/pull/50) - PR: Add versioneer for version string control based on git tags not on manual editing ([49](https://github.com/spyder-ide/loghub/issues/49))
* [PR 48](https://github.com/spyder-ide/loghub/pull/48) - PR: Update AppVeyor badge because of move to org account ([47](https://github.com/spyder-ide/loghub/issues/47))
* [PR 45](https://github.com/spyder-ide/loghub/pull/45) - PR: Break into smaller funcs
* [PR 44](https://github.com/spyder-ide/loghub/pull/44) - PR: Improve error handling ([40](https://github.com/spyder-ide/loghub/issues/40), [16](https://github.com/spyder-ide/loghub/issues/16))
* [PR 43](https://github.com/spyder-ide/loghub/pull/43) - PR: Add template for release grouped issue labels ([41](https://github.com/spyder-ide/loghub/issues/41))
* [PR 36](https://github.com/spyder-ide/loghub/pull/36) - PR: Update readme with latest CLI commands ([23](https://github.com/spyder-ide/loghub/issues/23))
* [PR 34](https://github.com/spyder-ide/loghub/pull/34) - PR: Add label grouping for issues  ([30](https://github.com/spyder-ide/loghub/issues/30))
* [PR 32](https://github.com/spyder-ide/loghub/pull/32) - PR: Feature/branch ([21](https://github.com/spyder-ide/loghub/issues/21))
* [PR 31](https://github.com/spyder-ide/loghub/pull/31) - PR: Refactor code
* [PR 29](https://github.com/spyder-ide/loghub/pull/29) - PR: Maintenance/tests ([2](https://github.com/spyder-ide/loghub/issues/2))
* [PR 20](https://github.com/spyder-ide/loghub/pull/20) - PR: Feature/templates ([10](https://github.com/spyder-ide/loghub/issues/10))
* [PR 19](https://github.com/spyder-ide/loghub/pull/19) - PR: Add support for all issues when no tag or no milestone is provided ([8](https://github.com/spyder-ide/loghub/issues/8))
* [PR 18](https://github.com/spyder-ide/loghub/pull/18) - PR: Add support for github tokens ([3](https://github.com/spyder-ide/loghub/issues/3))
* [PR 17](https://github.com/spyder-ide/loghub/pull/17) - PR: Add regex label filtering for issues and prs ([4](https://github.com/spyder-ide/loghub/issues/4))
* [PR 14](https://github.com/spyder-ide/loghub/pull/14) - add password prompt when user is specified and password is blank ([11](https://github.com/spyder-ide/loghub/issues/11))
* [PR 13](https://github.com/spyder-ide/loghub/pull/13) - PR: Fix bug on since keyword returning incorrect issues ([15](https://github.com/spyder-ide/loghub/issues/15))
* [PR 12](https://github.com/spyder-ide/loghub/pull/12) - Don't test macOS in Travis
* [PR 9](https://github.com/spyder-ide/loghub/pull/9) - Don't show "Issues" or "Pull requests" sections if there are no issues or PRs to show
* [PR 7](https://github.com/spyder-ide/loghub/pull/7) - Fix contents of temporary file
* [PR 6](https://github.com/spyder-ide/loghub/pull/6) - Remove extra empty lines
* [PR 5](https://github.com/spyder-ide/loghub/pull/5) - Fix PR_SHORT formatting.
* [PR 1](https://github.com/spyder-ide/loghub/pull/1) - Several fixes

In this release 29 pull requests were closed.
'''
    print([log])
    print([expected])
    assert log == expected


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_changelog():
    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE,
        branch='master',
        output_format='changelog')
    expected = '''## Version test-milestone (2016-12-05)

### Issues Closed

* [Issue 77](https://github.com/spyder-ide/loghub/issues/77) - Test empty body
* [Issue 26](https://github.com/spyder-ide/loghub/issues/26) - Test number 2
* [Issue 24](https://github.com/spyder-ide/loghub/issues/24) - Issue test

In this release 3 issues were closed.

### Pull Requests Merged

* [PR 25](https://github.com/spyder-ide/loghub/pull/25) - PR: Add tests folder

In this release 1 pull request was closed.
'''
    print([log])
    print([expected])
    assert log == expected


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_changelog_release():
    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE,
        branch='master',
        output_format='release')
    expected = '''## Version test-milestone (2016-12-05)

### Issues Closed

* Issue #77 - Test empty body
* Issue #26 - Test number 2
* Issue #24 - Issue test

In this release 3 issues were closed.

### Pull Requests Merged

* PR #25 - PR: Add tests folder

In this release 1 pull request was closed.
'''
    print([log])
    print([expected])
    assert log == expected


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_changelog_release_branch():
    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE,
        branch='test-branch',
        output_format='release')
    expected = '''## Version test-milestone (2016-12-05)

### Issues Closed

* Issue #77 - Test empty body
* Issue #26 - Test number 2
* Issue #24 - Issue test

In this release 3 issues were closed.

### Pull Requests Merged

* PR #85 - Test no repo in body
* PR #84 - Test no issue number in PR's body
* PR #33 - PR: Test change

In this release 3 pull requests were closed.
'''
    print([log])
    print([expected])
    assert log == expected


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
@patch('loghub.core.formatter.GitHubRepo')
def test_changelog_label_groups(gh_mock):
    gh_mock.return_value = gh_obj = MagicMock()
    gh_obj.milestone.return_value = {'closed_at': '2016-12-05'}
    issues = [{
        'loghub_label_names': ['type:bug'],
        'number': 1,
        'title': 'first issue',
        'html_url': 'a_url',
    }]
    prs = [{
        'loghub_label_names': ['type:bug'],
        'number': 2,
        'title': 'first pull request',
        'html_url': 'a_url',
        'pull_request': True,
        'body': '',
    }]
    gh_obj.issues.return_value = [JsonObject(x) for x in issues + prs]
    issue_label_groups = [{'label': 'type:bug', 'name': 'Bugs fixed'}]

    # group issues and PRs
    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE,
        branch='test-branch',
        issue_label_groups=issue_label_groups,
        pr_label_groups=issue_label_groups)
    expected = '''## Version test-milestone (2016-12-05)

#### Bugs fixed

* [Issue 1](https://github.com/spyder-ide/loghub/issues/1) - first issue
* [PR 2](https://github.com/spyder-ide/loghub/pull/2) - first pull request

In this release 1 issue and 1 pull request were closed.
'''
    print([log])
    print([expected])
    assert log == expected

    # group issues only, list PRs
    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE,
        branch='test-branch',
        issue_label_groups=issue_label_groups)
    expected = '''## Version test-milestone (2016-12-05)

### Issues Closed

#### Bugs fixed

* [Issue 1](https://github.com/spyder-ide/loghub/issues/1) - first issue

In this release 1 issue was closed.

### Pull Requests Merged

* [PR 2](https://github.com/spyder-ide/loghub/pull/2) - first pull request

In this release 1 pull request was closed.
'''
    print([log])
    print([expected])
    assert log == expected

    # group PRs only, list issues
    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE,
        branch='test-branch',
        pr_label_groups=issue_label_groups)
    expected = '''## Version test-milestone (2016-12-05)

### Issues Closed

* [Issue 1](https://github.com/spyder-ide/loghub/issues/1) - first issue

In this release 1 issue was closed.

### Pull Requests Merged

#### Bugs fixed

* [PR 2](https://github.com/spyder-ide/loghub/pull/2) - first pull request

In this release 1 pull request was closed.
'''
    print([log])
    print([expected])
    assert log == expected


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_changelog_issue_groups():
    issue_label_groups = [{'label': 'type:bug', 'name': 'Bugs fixed'}]
    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE,
        branch='test-branch',
        issue_label_groups=issue_label_groups)
    expected = '''## Version test-milestone (2016-12-05)

### Issues Closed

#### Bugs fixed

* [Issue 26](https://github.com/spyder-ide/loghub/issues/26) - Test number 2

In this release 1 issue was closed.

### Pull Requests Merged

* [PR 85](https://github.com/spyder-ide/loghub/pull/85) - Test no repo in body
* [PR 84](https://github.com/spyder-ide/loghub/pull/84) - Test no issue number in PR's body
* [PR 33](https://github.com/spyder-ide/loghub/pull/33) - PR: Test change

In this release 3 pull requests were closed.
'''
    print([log])
    print([expected])
    assert log == expected


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
@patch('loghub.core.formatter.GitHubRepo')
def test_no_comments(gh_mock):
    gh_mock.return_value = gh_obj = MagicMock()
    gh_obj.milestone.return_value = {'closed_at': '2016-12-05'}
    prs = [{
        'loghub_label_names': [],
        'number': 1,
        'title': 'pull request',
        'html_url': 'spyder-ide/loghub',
        'pull_request': True,
        'body': '<!--- Fixes #777 --->\nFixes #12',
    }]
    gh_obj.issues.return_value = [JsonObject(x) for x in prs]

    # group issues and PRs
    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE,
        branch='test-branch')
    expected = '''## Version test-milestone (2016-12-05)


### Pull Requests Merged

* [PR 1](https://github.com/spyder-ide/loghub/pull/1) - pull request ([12](https://github.com/spyder-ide/loghub/issues/12))

In this release 1 pull request was closed.
'''
    print([log])
    print([expected])
    assert log == expected


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_changelog_template():
    template = '''{%   for i in issues -%}
* Issue #{{ i['number'] }} - {{ i['title'] }}
{%   endfor %}'''
    desc, path = tempfile.mkstemp()
    with open(path, 'w') as f:
        f.write(template)

    log = create_changelog(
        repo=REPO,
        token=TEST_TOKEN,
        milestone=TEST_MILESTONE,
        template_file=path)
    expected = '* Issue #77 - Test empty body\n* Issue #26 - Test number 2\n* Issue #24 - Issue test\n'
    print([log])
    print([expected])
    assert log == expected


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_argument_parser_invalid():
    args = ['prog']
    with pytest.raises(SystemExit):
        with patch.object(sys, 'argv', args):
            parse_arguments(skip=True)


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_argument_parser_valid():
    args = [
        'prog', 'spyder-ide/loghub', '-ilg', 'type:bug', 'Bugs fixed', '-t',
        TEST_TOKEN
    ]
    with patch.object(sys, 'argv', args):
        options = parse_arguments()
    assert options.issue_label_groups == [['type:bug', 'Bugs fixed']]



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""ZenHub API."""

from __future__ import print_function

# Third party imports
import requests


class ZenHubError(Exception):
    pass


class ZenHub(object):
    """ZenHub API wrapper."""
    _HEADERS = {
        'Content-type': 'application/json',
        'User-Agent': 'ZenHub Python Client',
    }

    def __init__(self, token):
        """ZenHub API wrapper."""
        self._token = token
        self._endpoint = 'https://api.zenhub.com/p1/'
        self._session = requests.Session()

        # Setup
        self._session.headers.update(self._HEADERS)
        self._session.headers.update({'X-Authentication-Token': token})

    # --- Helpers
    @staticmethod
    def _parse_response_contents(response):
        """Parse response and convert to json if possible."""
        contents = {}

        status_code = response.status_code
        if status_code == 200:
            try:
                contents = response.json()                
            except Exception as err:
                print(err)
        elif status_code == 401:
            raise ZenHubError('Invalid token!')
        elif status_code == 403:
            raise ZenHubError('Reached request limit to the API. See API Limits.')
        elif status_code == 404:
            raise ZenHubError('Not found!')
        else:
            raise ZenHubError('Unknown error!')

        return contents

    def _make_url(self, url):
        """Create full api url."""
        return '{}{}'.format(self._endpoint, url)

    def _get(self, url):
        """Send GET request with given url."""
        response = self._session.get(self._make_url(url))
        return self._parse_response_contents(response)

    def releases(self, repo_id):
        """Provide list of releases for a given `repo_id`."""
        url = 'repositories/{repo_id}/reports/releases'
        return self._get(url.format(repo_id=repo_id))

    def issues(self, release_id):
        """Return a list of issues found in given `release_id`."""
        url = 'reports/release/{release_id}/issues'
        return self._get(url.format(release_id=release_id))



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Tests utils."""


class Issue:
    """Issue github mock."""

    def __init__(self, number, user='test-user', full_repo='foo/bar', body=None, is_pr=False):
        self.pull_request = is_pr
        self.number = number
        self.body = body or ''
        self.user = {'login': user, 'html_url': 'https://github.com/{}'.format(user)}

        if is_pr:
            self.html_url = 'https://github.com/{}/pull/{}'.format(full_repo, number)
        else:
            self.html_url = 'https://github.com/{}/issues/{}'.format(full_repo, number)

    def __repr__(self):
        return repr(self.__dict__)

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def get(self, name, default=None):
        if default is None:
            return self.__dict__.get(name)
        else:
            return self.__dict__.get(name, default)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Tests changelog labels."""

# Standard library imports
import os
import tempfile

# Third party imports
from mock import patch
import pytest

# Local imports
from loghub.core.labels import process_labels


# --- Mocks
# ----------------------------------------------------------------------------
class Label:

    def __init__(self, name, color):
        self.name = name
        self.color = color


class GHRepo:

    def __init__(self, *args, **kwargs):
        pass

    def labels(self):
        return [Label('foo', '#ff00ff'), Label('bar', '#0000ff')]

    def set_labels(self, labels):
        GHRepo._labels = labels


# --- Tests
# ----------------------------------------------------------------------------
def test_invalid_action_labels():
    _, tempfilepath = tempfile.mkstemp()

    with patch('loghub.core.labels.GitHubRepo', new=GHRepo) as ghrepo:
        with pytest.raises(ValueError):
            process_labels('', '', '', 'BOOM', 'monty-repo', tempfilepath)

        with open(tempfilepath, 'r') as fh:
            data = fh.read()

        assert data == ''


def test_get_labels():
    _, tempfilepath = tempfile.mkstemp()

    with patch('loghub.core.labels.GitHubRepo', new=GHRepo) as gh_repo:
        process_labels('', '', '', 'get', 'monty-repo', tempfilepath)

        with open(tempfilepath, 'r') as fh:
            data = fh.read()

        assert data == 'bar;#0000ff\nfoo;#ff00ff'


def test_set_labels_2_parts():
    _, tempfilepath = tempfile.mkstemp()

    with open(tempfilepath, 'w') as fh:
        data = fh.write('bar;#0000ff\nfoo;#ff00ff')

    with patch('loghub.core.labels.GitHubRepo', new=GHRepo) as gh_repo:
        process_labels('', '', '', 'update', 'monty-repo', tempfilepath)
        assert gh_repo._labels == [
            {'new_name': 'bar', 'old_name': 'bar', 'color': '#0000ff'},
            {'new_name': 'foo', 'old_name': 'foo', 'color': '#ff00ff'},
        ]


def test_set_labels_3_parts():
    _, tempfilepath = tempfile.mkstemp()

    with open(tempfilepath, 'w') as fh:
        data = fh.write('old_bar;bar;#0000ff\nold_foo;foo;#ff00ff')

    with patch('loghub.core.labels.GitHubRepo', new=GHRepo) as gh_repo:
        process_labels('', '', '', 'update', 'monty-repo', tempfilepath)
        assert gh_repo._labels == [
            {'new_name': 'bar', 'old_name': 'old_bar', 'color': '#0000ff'},
            {'new_name': 'foo', 'old_name': 'old_foo', 'color': '#ff00ff'},
        ]



#!/usr/bin/env python
# -*-coding: utf8 -*-
'''
GitHub API Python SDK. (Python >= 2.6)

Apache License

Michael Liao (askxuefeng@gmail.com)

Usage:

>>> gh = GitHub(username='githubpy', password='test-githubpy-1234')
>>> L = gh.users('githubpy').followers.get()
>>> L[0].id
470058
>>> L[0].login == u'michaelliao'
True
>>> x_ratelimit_remaining = gh.x_ratelimit_remaining
>>> x_ratelimit_limit = gh.x_ratelimit_limit
>>> x_ratelimit_reset = gh.x_ratelimit_reset
>>> L = gh.users('githubpy').following.get()
>>> L[0].url == u'https://api.github.com/users/michaelliao'
True
>>> L = gh.repos('githubpy')('testgithubpy').issues.get(state='closed', sort='created')
>>> L[0].title == u'sample issue for test'
True
>>> L[0].number
1
>>> I = gh.repos('githubpy')('testgithubpy').issues(1).get()
>>> I.url == u'https://api.github.com/repos/githubpy/testgithubpy/issues/1'
True
>>> gh = GitHub(username='githubpy', password='test-githubpy-1234')
>>> r = gh.repos('githubpy')('testgithubpy').issues.post(title='test create issue', body='just a test')
>>> r.title == u'test create issue'
True
>>> r.state == u'open'
True
>>> gh.repos.thisisabadurl.get()
Traceback (most recent call last):
    ...
ApiNotFoundError: https://api.github.com/repos/thisisabadurl
>>> gh.users('github-not-exist-user').followers.get()
Traceback (most recent call last):
    ...
ApiNotFoundError: https://api.github.com/users/github-not-exist-user/followers
'''
# yapf: disable
__version__ = '1.1.1'

try:
    # Python 2
    from urllib2 import build_opener, HTTPSHandler, Request, HTTPError
    from urllib import quote as urlquote
    from StringIO import StringIO
    from collections import Iterable

    def bytes(string, encoding=None):
        return str(string)
except:
    # Python 3
    from collections.abc import Iterable
    from urllib.request import build_opener, HTTPSHandler, HTTPError, Request
    from urllib.parse import quote as urlquote
    from io import StringIO

# Standard library imports
from datetime import datetime, timedelta, tzinfo
import base64
import hashlib
import hmac
import json
import mimetypes
import os
import re
import time
import urllib

TIMEOUT = 60

_URL = 'https://api.github.com'
_METHOD_MAP = dict(
    GET=lambda: 'GET',
    PUT=lambda: 'PUT',
    POST=lambda: 'POST',
    PATCH=lambda: 'PATCH',
    DELETE=lambda: 'DELETE')

DEFAULT_SCOPE = None
RW_SCOPE = 'user,public_repo,repo,repo:status,gist'


def _encode_params(kw):
    '''
    Encode parameters.
    '''
    args = []
    for k, v in kw.items():
        # If value is None or empty, ignore
        if v:
            try:
                # Python 2
                if not isinstance(v, [unicode, str]):
                    v = str(v)
                qv = v.encode('utf-8') if isinstance(v, unicode) else str(v)
            except:
                # Make sure all values are strings
                qv = str(v)
            args.append('%s=%s' % (k, urlquote(qv)))
    return '&'.join(args)


def _encode_json(obj):
    '''
    Encode object as json str.
    '''

    def _dump_obj(obj):
        if isinstance(obj, dict):
            return obj
        d = dict()
        for k in dir(obj):
            if not k.startswith('_'):
                d[k] = getattr(obj, k)
        return d

    return json.dumps(obj, default=_dump_obj)


def _parse_json(jsonstr):
    def _obj_hook(pairs):
        o = JsonObject()
        for k, v in pairs.items():
            o[str(k)] = v
        return o

    return json.loads(jsonstr, object_hook=_obj_hook)


class _Executable(object):
    def __init__(self, _gh, _method, _path):
        self._gh = _gh
        self._method = _method
        self._path = _path

    def __call__(self, **kw):
        return self._gh._http(self._method, self._path, **kw)

    def __str__(self):
        return '_Executable (%s %s)' % (self._method, self._path)

    __repr__ = __str__


class _Callable(object):
    def __init__(self, _gh, _name):
        self._gh = _gh
        self._name = _name

    def __call__(self, *args):
        if len(args) == 0:
            return self
        name = '%s/%s' % (self._name, '/'.join([str(arg) for arg in args]))
        return _Callable(self._gh, name)

    def __getattr__(self, attr):
        if attr == 'get':
            return _Executable(self._gh, 'GET', self._name)
        if attr == 'put':
            return _Executable(self._gh, 'PUT', self._name)
        if attr == 'post':
            return _Executable(self._gh, 'POST', self._name)
        if attr == 'patch':
            return _Executable(self._gh, 'PATCH', self._name)
        if attr == 'delete':
            return _Executable(self._gh, 'DELETE', self._name)
        name = '%s/%s' % (self._name, attr)
        return _Callable(self._gh, name)

    def __str__(self):
        return '_Callable (%s)' % self._name

    __repr__ = __str__


class GitHub(object):
    '''
    GitHub client.
    '''

    def __init__(self,
                 username=None,
                 password=None,
                 access_token=None,
                 client_id=None,
                 client_secret=None,
                 redirect_uri=None,
                 scope=None):
        self.x_ratelimit_remaining = (-1)
        self.x_ratelimit_limit = (-1)
        self.x_ratelimit_reset = (-1)
        self._authorization = None
        if username and password:
            # roundabout hack for Python 3
            userandpass = base64.b64encode(
                bytes('%s:%s' % (username, password), 'utf-8'))
            userandpass = userandpass.decode('ascii')
            self._authorization = 'Basic %s' % userandpass
        elif access_token:
            self._authorization = 'token %s' % access_token
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._scope = scope

    def authorize_url(self, state=None):
        '''
        Generate authorize_url.

        >>> GitHub(client_id='3ebf94c5776d565bcf75').authorize_url()
        'https://github.com/login/oauth/authorize?client_id=3ebf94c5776d565bcf75'
        '''
        if not self._client_id:
            raise ApiAuthError('No client id.')
        kw = dict(client_id=self._client_id)
        if self._redirect_uri:
            kw['redirect_uri'] = self._redirect_uri
        if self._scope:
            kw['scope'] = self._scope
        if state:
            kw['state'] = state
        return 'https://github.com/login/oauth/authorize?%s' % _encode_params(
            kw)

    def get_access_token(self, code, state=None):
        '''
        In callback url: http://host/callback?code=123&state=xyz

        use code and state to get an access token.        
        '''
        kw = dict(
            client_id=self._client_id,
            client_secret=self._client_secret,
            code=code)
        if self._redirect_uri:
            kw['redirect_uri'] = self._redirect_uri
        if state:
            kw['state'] = state
        opener = build_opener(HTTPSHandler)
        request = Request(
            'https://github.com/login/oauth/access_token',
            data=_encode_params(kw))
        request.get_method = _METHOD_MAP['POST']
        request.add_header('Accept', 'application/json')
        try:
            response = opener.open(request, timeout=TIMEOUT)
            r = _parse_json(response.read())
            if 'error' in r:
                raise ApiAuthError(str(r.error))
            return str(r.access_token)
        except HTTPError as e:
            raise ApiAuthError('HTTPError when get access token')

    def __getattr__(self, attr):
        return _Callable(self, '/%s' % attr)

    def _http(self, _method, _path, **kw):
        data = None
        params = None
        if _method == 'GET' and kw:
            _path = '%s?%s' % (_path, _encode_params(kw))
        if _method in ['POST', 'PATCH', 'PUT']:
            data = bytes(_encode_json(kw), 'utf-8')
        url = '%s%s' % (_URL, _path)
        opener = build_opener(HTTPSHandler)
        request = Request(url, data=data)
        request.get_method = _METHOD_MAP[_method]
        if self._authorization:
            request.add_header('Authorization', self._authorization)
        if _method in ['POST', 'PATCH', 'PUT']:
            request.add_header('Content-Type',
                               'application/x-www-form-urlencoded')

        class Resp():
            code = None

        resp = Resp()
        req = None

        try:
            response = opener.open(request, timeout=TIMEOUT)
            is_json = self._process_resp(response.headers)
            if is_json:
                return _parse_json(response.read().decode('utf-8'))
        except HTTPError as e:
            is_json = self._process_resp(e.headers)
            if is_json:
                json = _parse_json(e.read().decode('utf-8'))
            else:
                json = e.read().decode('utf-8')
            req = JsonObject(method=_method, url=url)
            resp = JsonObject(code=e.code, json=json)
        finally:
            if resp.code == 404:
                raise ApiNotFoundError(url, req, resp)
            elif req:
                raise ApiError(url, req, resp)

    def _process_resp(self, headers):
        is_json = False
        if headers:
            for k in headers:
                h = k.lower()
                if h == 'x-ratelimit-remaining':
                    self.x_ratelimit_remaining = int(headers[k])
                elif h == 'x-ratelimit-limit':
                    self.x_ratelimit_limit = int(headers[k])
                elif h == 'x-ratelimit-reset':
                    self.x_ratelimit_reset = int(headers[k])
                elif h == 'content-type':
                    is_json = headers[k].startswith('application/json')
        return is_json


class JsonObject(dict):
    '''
    general json object that can bind any fields but also act as a dict.
    '''

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(r"'Dict' object has no attribute '%s'" % key)

    def __setattr__(self, attr, value):
        self[attr] = value


class ApiError(Exception):
    def __init__(self, url, request, response):
        super(ApiError, self).__init__(url)
        self.request = request
        self.response = response


class ApiAuthError(ApiError):
    def __init__(self, msg):
        super(ApiAuthError, self).__init__(msg, None, None)


class ApiNotFoundError(ApiError):
    pass


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Tests changelog labels."""

# Standard library imports
import tempfile

# Third party imports
from mock import patch

# Local imports
from loghub.core.config import load_config


def test_load_config_empty():
    _, tempfilepath = tempfile.mkstemp()

    with patch('loghub.core.config.user_config_path',
                return_value=tempfilepath) as method:
        config = load_config()
        assert config.sections() == []


def test_load_config():
    _, tempfilepath = tempfile.mkstemp()
    with open(tempfilepath, 'w') as fh:
        fh.write('[github]\ntoken=1\n[zenhub]\ntoken=2')

    with patch('loghub.core.config.user_config_path',
                return_value=tempfilepath) as method:
        config = load_config()
        assert config.sections() == ['github', 'zenhub']
        assert config.get('github', 'token') == '1'
        assert config.get('zenhub', 'token') == '2'


def test_load_config_with_path():
    _, tempfilepath = tempfile.mkstemp()
    with open(tempfilepath, 'w') as fh:
        fh.write('[github]\ntoken=1\n[zenhub]\ntoken=2')

    config = load_config(tempfilepath)
    assert config.sections() == ['github', 'zenhub']
    assert config.get('github', 'token') == '1'
    assert config.get('zenhub', 'token') == '2'



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Get/Create github labels from a text file."""

# yapf: disable

# Local imports
from loghub.core.repo import GitHubRepo

# yapf: enable


def process_labels(username, password, token, action, repo, filename):
    """Get or update labels on a given repository."""
    # Separator
    s = ';'

    # Instantiate Github API
    gh = GitHubRepo(
        username=username,
        password=password,
        token=token,
        repo=repo, )

    labels = gh.labels()

    if action == 'get':
        print('Getting labels from {0}\n'.format(repo))
        labels = sorted((l.name, l.color) for l in labels)
        data = ''.join('{0}{1}{2}\n'.format(l, s, c) for (l, c) in labels)

        with open(filename, 'wt') as f:
            f.write(data[:-1])

    elif action == 'update':
        print('Updating labels on {0}\n'.format(repo))
        with open(filename, 'r') as f:
            data = f.read()

        lines = [line for line in data.split('\n') if line]
        labels = []
        for line in lines:
            parts = [p for p in line.strip().split(s)]
            if len(parts) == 3:
                old_name, new_name, color = parts
            else:
                new_name, color = parts
                old_name = new_name

            label_dict = {
                'new_name': new_name,
                'old_name': old_name,
                'color': color
            }
            labels.append(label_dict)

        gh.set_labels(labels)
    else:
        raise ValueError('Invalid action!')



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Create github labels from a text file."""

# yapf: disable

# Standard library imports
import argparse

# Local imports
from loghub.cli.common import add_common_parser_args, parse_password_check_repo
from loghub.core.labels import process_labels

# yapf: enable


def main():
    """Main script."""
    parse_arguments(skip=False)


def parse_arguments(skip=False):
    """Parse argument for label creator utility."""
    # Get command-line arguments
    parser = argparse.ArgumentParser()

    parser = add_common_parser_args(parser)
    parser.add_argument(
        '-a',
        '--action',
        help='Action to take',
        type=str,
        choices=['get', 'update'],
        default='get',
        nargs='?')
    parser.add_argument(
        '-f',
        '--filename',
        help='File for storing labels',
        type=str,
        default='labels.txt')

    options = parser.parse_args()

    username = options.username
    password = parse_password_check_repo(options)

    if not skip:
        process_labels(
            username,
            password,
            options.token,
            options.action,
            options.repository,
            options.filename, )

    return options


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Templates for output formatting."""

# Standard library imports
import os

HERE = os.path.dirname(os.path.realpath(__file__))
CHANGELOG_TEMPLATE_PATH = os.path.join(HERE, 'changelog.txt')
CHANGELOG_GROUPS_TEMPLATE_PATH = os.path.join(HERE,
                                              'changelog_label_groups.txt')
CHANGELOG_ISSUE_GROUPS_TEMPLATE_PATH = os.path.join(
    HERE, 'changelog_issue_groups.txt')
CHANGELOG_PR_GROUPS_TEMPLATE_PATH = os.path.join(HERE,
                                                 'changelog_pr_groups.txt')
RELEASE_TEMPLATE_PATH = os.path.join(HERE, 'release.txt')
RELEASE_GROUPS_TEMPLATE_PATH = os.path.join(HERE, 'release_groups.txt')
RELEASE_GROUPS_TEMPLATE_PATH = os.path.join(HERE, 'release_label_groups.txt')
RELEASE_ISSUE_GROUPS_TEMPLATE_PATH = os.path.join(HERE,
                                                  'release_issue_groups.txt')
RELEASE_PR_GROUPS_TEMPLATE_PATH = os.path.join(HERE, 'release_pr_groups.txt')



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Build a list of issues and pull requests per Github milestone."""

# yapf: disable

from __future__ import print_function

# Standard library imports
import argparse
import sys
import textwrap

# Local imports
from loghub.cli.common import add_common_parser_args, parse_password_check_repo
from loghub.core.config import load_config
from loghub.core.formatter import create_changelog

# yapf: enable

PY2 = sys.version[0] == '2'


def check_github_deprecation(username='', password=''):
    """
    Inform users that username and password are deprecated by Github API.
    """
    text = (
        'Deprecation Notice: GitHub will discontinue password '
        'authentication to the API. You must now authenticate '
        'to the GitHub API with an API token, such as an OAuth '
        'access token, GitHub App installation access token, or '
        'personal access token, depending on what you need to do '
        'with the token. Password authentication to the API will be '
        'removed on November 13, 2020. For more information, including '
        'scheduled brownouts, see: '
        'https://developer.github.com/changes/'
        '2020-02-14-deprecating-password-auth/         '
        'To create a new token go to: https://github.com/settings/tokens'
    )
    if username or password:
        print('\n' + '\n'.join(textwrap.wrap(text, 79)) + '\n')


def main():
    """Main script."""
    parse_arguments(skip=False)


def create_label_groups(groups):
    """Create info dictionaries for label groups."""
    group_dicts = []
    if groups:
        for item in groups:
            dic = {}
            if len(item) == 1:
                dic['label'] = item[0]
                dic['name'] = item[0]
            elif len(item) == 2:
                dic['label'] = item[0]
                dic['name'] = item[1]
            else:
                raise ValueError('Label group takes 1 or 2 arguments')
            group_dicts.append(dic)
    return group_dicts


def parse_arguments(skip=False):
    """Parse CLI arguments."""
    parser = argparse.ArgumentParser(
        description='Script to print the list of issues and pull requests '
        'closed in a given milestone, tag including additional filtering '
        'options.')

    # Get common parser arguments
    parser = add_common_parser_args(parser)

    parser.add_argument(
        '-m',
        '--milestone',
        action="store",
        dest="milestone",
        default='',
        help="Github milestone to get issues and pull requests for")
    parser.add_argument(
        '-zr',
        '--zenhub-release',
        action="store",
        dest="zenhub_release",
        default='',
        help="Zenhub release to get issues and pull requests for")
    parser.add_argument(
        '-st',
        '--since-tag',
        action="store",
        dest="since_tag",
        default='',
        help="Github issues and pull requests since tag")
    parser.add_argument(
        '-ut',
        '--until-tag',
        action="store",
        dest="until_tag",
        default='',
        help="Github issues and pull requests until tag")
    parser.add_argument(
        '-b',
        '--branch',
        action="store",
        dest="branch",
        default='',
        help="Github base branch for merged PRs")
    parser.add_argument(
        '-ilg',
        '--issue-label-group',
        action="append",
        nargs='+',
        default=[],
        dest="issue_label_groups",
        help="Groups the generated issues by the specified label. This option"
        "takes 1 or 2 arguments, where the first one is the label to "
        "match and the second one is the label to print on the final"
        "output")
    parser.add_argument(
        '-plg',
        '--pr-label-group',
        action="append",
        nargs='+',
        default=[],
        dest="pr_label_groups",
        help="Groups the generated PRs by the specified label. This option"
        "takes 1 or 2 arguments, where the first one is the label to "
        "match and the second one is the label to print on the final"
        "output")
    parser.add_argument(
        '-lg',
        '--label-group',
        action="append",
        nargs='+',
        default=[],
        dest="label_groups",
        help="Groups the generated issues and PRs by the specified label. "
        "This option takes 1 or 2 arguments, where the first one is the "
        "label to match and the second one is the label to print on "
        "the final output")
    parser.add_argument(
        '-ilr',
        '--issue-label-regex',
        action="store",
        dest="issue_label_regex",
        default='',
        help="Label issue filter using a regular expression filter")
    parser.add_argument(
        '-plr',
        '--pr-label-regex',
        action="store",
        dest="pr_label_regex",
        default='',
        help="Label pull request filter using a regular expression filter")
    parser.add_argument(
        '-f',
        '--format',
        action="store",
        dest="output_format",
        default='changelog',
        help="Format for print, either 'changelog' (for "
        "Changelog.md file) or 'release' (for the Github "
        "Releases page). Default is 'changelog'. The "
        "'release' option doesn't generate Markdown "
        "hyperlinks.")
    parser.add_argument(
        '--template',
        action="store",
        dest="template",
        default='',
        help="Use a custom Jinja2 template file ")
    parser.add_argument(
        '--batch',
        action="store",
        dest="batch",
        default=None,
        choices=['milestones', 'tags'],
        help="Run loghub for all milestones or all tags")
    parser.add_argument(
        '--no-prs',
        action="store_false",
        dest="show_prs",
        default=True,
        help="Run loghub without any pull requests output")
    parser.add_argument(
        '--no-related-prs',
        action="store_false",
        dest="show_related_prs",
        default=True,
        help="Do not display related prs on issues")
    parser.add_argument(
        '--no-related-issues',
        action="store_false",
        dest="show_related_issues",
        default=True,
        help="Do not display related issues on prs")

    options = parser.parse_args()

    milestone = options.milestone
    zenhub_release = options.zenhub_release
    batch = options.batch

    # Check if milestone, release or tag given
    if not batch:
        if not milestone and not zenhub_release and not options.since_tag and not options.until_tag:
            print('\nLOGHUB: Querying all issues\n')
        elif milestone and not zenhub_release and not options.since_tag and not options.until_tag:
            print('\nLOGHUB: Querying issues for milestone "{0}"'
                  '\n'.format(milestone))
        elif zenhub_release and not milestone and not options.since_tag and not options.until_tag:
            print('\nLOGHUB: Querying issues for zenhub release "{0}"'
                  '\n'.format(zenhub_release))
        elif options.since_tag and not options.until_tag and not milestone and not zenhub_release:
            print('\nLOGHUB: Querying issues since tag "{0}"'
                  '\n'.format(options.since_tag))
        elif options.since_tag and options.until_tag and not milestone and not zenhub_release:
            print('\nLOGHUB: Querying issues since tag "{0}" until tag "{1}"'
                  '\n'.format(options.since_tag, options.until_tag))
        else:
            print('Invalid set of options!')
            sys.exit(1)
    elif batch and any([
            bool(options.since_tag), bool(options.until_tag),
            bool(milestone), bool(zenhub_release),
    ]):
        print('LOGHUB: When using batch mode no tags or milestone '
              ' or zenhub release arguments are allowed.\n')
        sys.exit(1)

    # Ask for password once input is valid
    username = options.username
    password = parse_password_check_repo(options)
    try:
        issue_label_groups = options.label_groups + \
                             options.issue_label_groups
        new_issue_label_groups = create_label_groups(issue_label_groups)
    except ValueError:
        print('LOGHUB: Issue label group takes 1 or 2 arguments\n')
        sys.exit(1)

    try:
        pr_label_groups = options.label_groups + options.pr_label_groups
        new_pr_label_groups = create_label_groups(pr_label_groups)
    except ValueError:
        print('LOGHUB: PR label group takes 1 or 2 arguments\n')
        sys.exit(1)

    # Inform users that username and password are deprecated by Github API
    check_github_deprecation(options.username, options.password)

    github_token = options.token
    zenhub_token = options.zenhub_token

    # Check for configuration file in home folder
    config = load_config()
    if config:
        try:
            github_token = config.get('github', 'token')
        except Exception:
            github_token = options.token

        try:
            zenhub_token = config.get('zenhub', 'token')
        except Exception:
            zenhub_token = options.zenhub_token

    if not skip:
        create_changelog(
            repo=options.repository,
            username=username,
            password=password,
            token=github_token,
            milestone=milestone,
            zenhub_release=zenhub_release,
            zenhub_token=zenhub_token,
            since_tag=options.since_tag,
            until_tag=options.until_tag,
            branch=options.branch,
            issue_label_regex=options.issue_label_regex,
            pr_label_regex=options.pr_label_regex,
            output_format=options.output_format,
            template_file=options.template,
            issue_label_groups=new_issue_label_groups,
            pr_label_groups=new_pr_label_groups,
            batch=batch,
            show_prs=options.show_prs,
            show_related_prs=options.show_related_prs,
            show_related_issues=options.show_related_issues,
        )

    return options


if __name__ == '__main__':  # yapf: disable
    main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Tests github repo."""

# Standard library imports
import os

# Third party imports
import pytest

# Local imports
from loghub.core.repo import GitHubRepo
from loghub.external.github import ApiError

REPO = 'spyder-ide/loghub'
TEST_TOKEN = os.environ.get('TEST_TOKEN', '').replace('x', '')
TEST_USER = os.environ.get('TEST_USER', '').replace('x', '')
TEST_PASS = os.environ.get('TEST_CODE', '').replace('x', '')
TEST_MILESTONE = 'test-milestone'
TEST_TAG = 'v0.1.2'
NOT_ON_CI = os.environ.get('CIRCLECI') != 'true'


# --- Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture
def gh_repo():
    return GitHubRepo(token=TEST_TOKEN, repo=REPO)


# --- Tests
# -----------------------------------------------------------------------------
@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_valid_user_password():
    gh = GitHubRepo(username=TEST_USER, password=TEST_PASS, repo=REPO)
    assert bool(gh.milestone(TEST_MILESTONE))


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_invalid_user_password():
    with pytest.raises(SystemExit):
        gh = GitHubRepo(
            username='invalid-user',
            password='invalid-password',
            repo=REPO, )
        gh.milestone(TEST_MILESTONE)


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_valid_token():
    gh = GitHubRepo(token=TEST_TOKEN, repo=REPO)
    gh.milestone(TEST_MILESTONE)


def test_invalid_token():
    with pytest.raises(SystemExit):
        gh = GitHubRepo(token='this-is-an-invalid-token', repo=REPO)
        gh.milestone(TEST_MILESTONE)


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_tags(gh_repo):
    tags = gh_repo.tags()
    titles = [tag['ref'].replace('refs/tags/', '') for tag in tags]
    assert TEST_TAG in titles


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_valid_tag(gh_repo):
    tag = gh_repo.tag(TEST_TAG)
    assert bool(tag)


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_invalid_tag(gh_repo):
    with pytest.raises(SystemExit):
        gh_repo.tag('invalid-tag')


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_milestones(gh_repo):
    milestones = gh_repo.milestones()
    titles = [milestone['title'] for milestone in milestones]
    assert TEST_MILESTONE in titles


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_valid_milestone(gh_repo):
    assert bool(gh_repo.milestone(TEST_MILESTONE))


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_invalid_milestone(gh_repo):
    with pytest.raises(SystemExit):
        gh_repo.milestone('invalid-milestone')


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_issues(gh_repo):
    issues = gh_repo.issues(milestone=TEST_MILESTONE, state='closed')
    assert len(issues) == 7


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_pr_merged(gh_repo):
    assert gh_repo.is_merged(25)


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_pr_closed(gh_repo):
    assert not gh_repo.is_merged(22)


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_pr_valid(gh_repo):
    pr = gh_repo.pr(32)
    assert bool(pr)


@pytest.mark.skipif(NOT_ON_CI, reason='test on ci server only')
def test_pr_invalid(gh_repo):
    with pytest.raises(ApiError):
        gh_repo.pr(1000000)


def test_dates():
    date = GitHubRepo.str_to_date('2016-10-10T08:08:08Z')
    assert date.year == 2016
    assert date.month == 10
    assert date.day == 10
    assert date.hour == 8
    assert date.minute == 8
    assert date.second == 8



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Changelog generator based on github milestones or tags."""

__version__ = '0.5.1'


def _to_version_info(version):
    """Convert a version string to a number and string tuple."""
    parts = []
    for part in version.split('.'):
        try:
            part = int(part)
        except ValueError:
            pass

        parts.append(part)

    return tuple(parts)


VERSION_INFO = _to_version_info(__version__)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Basic config parser."""

# Third party imports
import os
import sys

# Constants
PY2 = sys.version[0] == '2'
if PY2:
    import ConfigParser as configparser
else:
    import configparser


def user_config_path():
    """Return the path of the Loghub user configuration file."""
    return os.path.expanduser('~/.loghubrc')


def load_config(path=None):
    """Load configutration file."""
    config_path = path or user_config_path()
    config = configparser.ConfigParser()
    if os.path.isfile(config_path):
        config.read(config_path)

    return config



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Common CLI utilities."""

# yapf: disable

from __future__ import print_function

# Standard library imports
import getpass
import sys

# yapf: enable


def add_common_parser_args(parser):
    """Parse CLI common login and repo arguments."""
    parser.add_argument(
        'repository',
        help="Repository name to generate the Changelog for, in the form "
        "user/repo or org/repo (e.g. spyder-ide/spyder)")
    parser.add_argument(
        '-u',
        '--username',
        action="store",
        dest="username",
        default='',
        help="Github user name")
    parser.add_argument(
        '-p',
        '--password',
        action="store",
        dest="password",
        default='',
        help="Github user password")
    parser.add_argument(
        '-t',
        '--token',
        action="store",
        dest="token",
        default='',
        help="Github access token")
    parser.add_argument(
        '-zt',
        '--zenhub-token',
        action="store",
        dest="zenhub_token",
        default='',
        help="Zenhub access token")

    return parser


def parse_password_check_repo(options):
    """Check password and prompt if missing and check repo is provided."""
    if options.username and not options.password:
        password = getpass.getpass()
    else:
        password = options.password

    # Check if repo given
    if not options.repository:
        print('LOGHUB: Please define a repository name to this script. '
              'See its help')
        sys.exit(1)
    return password



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Setup script for loghub."""

# Standard library imports
import ast
import os

# Third party imports
from setuptools import find_packages, setup

# Constants
HERE = os.path.abspath(os.path.dirname(__file__))


def get_version(module='loghub'):
    """Get version."""
    with open(os.path.join(HERE, module, '__init__.py'), 'r') as f:
        data = f.read()

    lines = data.split('\n')
    for line in lines:
        if line.startswith('__version__'):
            version = ast.literal_eval(line.split('=')[-1].strip())
            break

    return version


def get_description():
    """Get long description."""
    with open(os.path.join(HERE, 'README.md'), 'r') as f:
        data = f.read()
    return data


REQUIREMENTS = [
    'jinja2',
    'requests',
]

setup(
    name='loghub',
    version=get_version(),
    keywords=["github zenhub changelog milestone releases"],
    url='https://github.com/spyder-ide/loghub',
    license='MIT',
    author='Carlos Cordoba',
    author_email='ccordoba12@gmail.com',
    maintainer='Carlos Cordoba',
    maintainer_email='ccordoba12@gmail.com',
    description='Generate changelogs based on Github milestones or tags',
    long_description=get_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    package_data={'loghub.templates': ['*.txt']},
    install_requires=REQUIREMENTS,
    entry_points={
        'console_scripts': [
            'loghub = loghub.cli.main:main',
            'loghub-labels = loghub.cli.label_creator:main',
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Tests formatter."""

# Standard library imports
import os

# Third party imports
import pytest

# Local imports
from loghub.core.formatter import filter_issues_fixed_by_prs
from loghub.tests.utils import Issue


# --- Tests
# -----------------------------------------------------------------------------
def test_filter_issues_fixed_by_prs_no_show():
    issues = [
        Issue(number=34),
        Issue(number=35),
    ]
    prs = [
        Issue(body='Fixes #34', is_pr=True, number=45),
        Issue(body='Closes #35', is_pr=True, number=46),
    ]

    new_issues, new_prs = filter_issues_fixed_by_prs(
        issues, prs, show_related_prs=False, show_related_issues=False
    )
    assert not all(bool(issue['loghub_related_pulls']) for issue in new_issues)
    assert not all(bool(pr['loghub_related_issues']) for pr in new_prs)


def test_filter_issues_fixed_by_prs_no_show_issues():
    issues = [
        Issue(number=34),
        Issue(number=35),
    ]
    prs = [
        Issue(body='Fixes #34', is_pr=True, number=45),
        Issue(body='Closes #35', is_pr=True, number=46),
    ]

    new_issues, new_prs = filter_issues_fixed_by_prs(
        issues, prs, show_related_prs=True, show_related_issues=False
    )
    assert all(bool(issue['loghub_related_pulls']) for issue in new_issues)
    assert not all(bool(pr['loghub_related_issues']) for pr in new_prs)


def test_filter_issues_fixed_by_prs_no_show_pulls():
    issues = [
        Issue(number=34),
        Issue(number=35),
    ]
    prs = [
        Issue(body='Fixes #34', is_pr=True, number=45),
        Issue(body='Closes #35', is_pr=True, number=46),
    ]

    new_issues, new_prs = filter_issues_fixed_by_prs(
        issues, prs, show_related_prs=False, show_related_issues=True
    )
    assert not all(bool(issue['loghub_related_pulls']) for issue in new_issues)
    assert all(bool(pr['loghub_related_issues']) for pr in new_prs)


def test_filter_issues_fixed_by_prs_show():
    issues = [
        Issue(number=34),
        Issue(number=35),
    ]
    prs = [
        Issue(body='Fixes #34', is_pr=True, number=45),
        Issue(body='Closes #35', is_pr=True, number=46),
    ]

    new_issues, new_prs = filter_issues_fixed_by_prs(
        issues, prs, show_related_prs=True, show_related_issues=True
    )
    assert all(bool(issue['loghub_related_pulls']) for issue in new_issues)
    assert all(bool(pr['loghub_related_issues']) for pr in new_prs)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Github repo wrapper."""

from __future__ import print_function

# Standard library imports
import datetime
import sys
import time

# Local imports
from loghub.external.github import ApiError, ApiNotFoundError, GitHub


class GitHubRepo(object):
    """Github repository wrapper."""

    def __init__(self, username=None, password=None, token=None, repo=None):
        """Github repository wrapper."""
        self._username = username
        self._password = password
        self._token = token

        self.gh = GitHub(
            username=username,
            password=password,
            access_token=token, )
        repo_organization, repo_name = repo.split('/')
        self._repo_organization = repo_organization
        self._repo_name = repo_name
        self.repo = self.gh.repos(repo_organization)(repo_name)

        # Check username and repo name
        self._check_user()
        self._check_repo_name()

    def _check_user(self):
        """Check if the supplied username is valid."""
        try:
            self.gh.users(self._repo_organization).get()
        except ApiNotFoundError:
            print('LOGHUB: Organization/user `{}` seems to be '
                  'invalid.\n'.format(self._repo_organization))
            sys.exit(1)
        except ApiError:
            self._check_rate()
            print('LOGHUB: The credentials seems to be invalid!\n')
            sys.exit(1)

    def _check_repo_name(self):
        """Check if the supplied repository exists."""
        try:
            self.repo.get()
        except ApiNotFoundError:
            print('LOGHUB: Repository `{0}` for organization/username `{1}` '
                  'seems to be invalid.\n'.format(self._repo_name,
                                                  self._repo_organization))
            sys.exit(1)
        except ApiError:
            self._check_rate()

    def _check_rate(self):
        """Check and handle if api rate limit has been exceeded."""
        if self.gh.x_ratelimit_remaining == 0:
            reset_struct = time.gmtime(self.gh.x_ratelimit_reset)
            reset_format = time.strftime('%Y/%m/%d %H:%M', reset_struct)
            print('LOGHUB: GitHub API rate limit exceeded!')
            print('LOGHUB: GitHub API rate limit resets on '
                  '{}'.format(reset_format))
            if not self._username and not self._password or not self._token:
                print('LOGHUB: Try running loghub with user/password or '
                      'a valid token.\n')
            sys.exit(1)

    def _filter_milestone(self, issues, milestone):
        """Filter out all issues in milestone."""
        if milestone:
            for issue in issues[:]:
                milestone_data = issue.get('milestone', {})
                if milestone_data:
                    issue_milestone_title = milestone_data.get('title')
                else:
                    issue_milestone_title = ''

                if issue_milestone_title != milestone:
                    issues.remove(issue)
        return issues

    def _filter_since(self, issues, since):
        """Filter out all issues before `since` date."""
        if since:
            since_date = self.str_to_date(since)
            for issue in issues[:]:
                close_date = self.str_to_date(issue['closed_at'])
                if close_date < since_date and issue in issues:
                    issues.remove(issue)
        return issues

    def _filter_until(self, issues, until):
        """Filter out all issues after `until` date."""
        if until:
            until_date = self.str_to_date(until)
            for issue in issues[:]:
                close_date = self.str_to_date(issue['closed_at'])
                if close_date > until_date and issue in issues:
                    issues.remove(issue)
        return issues

    def _filter_by_branch(self, issues, issue, branch):
        """Filter prs by the branch they were merged into."""
        number = issue['number']

        if not self.is_merged(number) and issue in issues:
            issues.remove(issue)

        if branch:
            # Get PR info and get base branch
            pr_data = self.pr(number)
            base_ref = pr_data['base']['ref']

            if base_ref != branch and issue in issues:
                issues.remove(issue)

        return issues

    def _filer_closed_prs(self, issues, branch):
        """Filter out closed PRs."""
        for issue in issues[:]:
            pr = issue.get('pull_request', '')

            # Add label names inside additional key
            issue['loghub_label_names'] = [
                l['name'] for l in issue.get('labels')
            ]

            if pr:
                issues = self._filter_by_branch(issues, issue, branch)

        return issues

    def tags(self):
        """Return all tags."""
        self._check_rate()
        return self.repo('git')('refs')('tags').get()

    def tag(self, tag_name):
        """Get tag information."""
        self._check_rate()
        refs = self.repo('git')('refs')('tags').get()
        sha = -1

        tags = []
        for ref in refs:
            ref_name = 'refs/tags/{tag}'.format(tag=tag_name)
            if 'object' in ref and ref['ref'] == ref_name:
                sha = ref['object']['sha']
            tags.append(ref['ref'].split('/')[-1])

        if sha == -1:
            print("LOGHUB: You didn't pass a valid tag name!")
            print('LOGHUB: The available tags are: {0}\n'.format(tags))
            sys.exit(1)

        try:
            tag = self.repo('git')('tags')(sha).get()
        except ApiNotFoundError:
            print("\nLOGHUB: The Git tags API only supports annotated tag "
                  "objects, not lightweight tags.\n")
            sys.exit(1)

        return tag

    def labels(self):
        """Return labels for the repo."""
        self._check_rate()
        return self.repo.labels.get()

    def set_labels(self, labels):
        """Return labels for the repo."""
        self._check_rate()
        for label in labels:
            new_name = label['new_name']
            old_name = label['old_name']
            color = label['color']
            try:
                self.repo.labels(old_name).patch(name=new_name, color=color)
                print('Updated label: "{0}" -> "{1}" (#{2})'.format(
                    old_name, new_name, color))
            except ApiError:
                try:
                    self.repo.labels.post(name=new_name, color=color)
                    print('Created label: "{0}" (#{1})'.format(new_name,
                                                               color))
                except ApiError:
                    print('\nLabel "{0}" already exists!'.format(new_name))

    def milestones(self):
        """Return all milestones."""
        self._check_rate()
        return self.repo.milestones.get(state='all', direction='desc')

    def milestone(self, milestone_title):
        """Return milestone with given title."""
        self._check_rate()
        milestones = self.milestones()
        milestone_number = -1

        milestone_titles = [milestone['title'] for milestone in milestones]
        for milestone in milestones:
            if milestone['title'] == milestone_title:
                milestone_number = milestone['number']
                break

        if milestone_number == -1:
            print("LOGHUB: You didn't pass a valid milestone name!")
            print('LOGHUB: The available milestones are: {0}\n'
                  ''.format(milestone_titles))
            sys.exit(1)

        return milestone

    def pr(self, pr_number):
        """Get PR information."""
        self._check_rate()
        return self.repo('pulls')(str(pr_number)).get()

    def issue(self, issue_number):
        """Get a specific issue number from repo."""
        # /repos/:owner/:repo/issues/:issue_number
        self._check_rate()
        return self.repo.issues(str(issue_number)).get()

    def issues(self,
               milestone=None,
               state=None,
               assignee=None,
               creator=None,
               mentioned=None,
               labels=None,
               sort=None,
               direction=None,
               since=None,
               until=None,
               branch=None,
               cache=False,
               base_issues=None):
        """Return Issues and Pull Requests."""
        self._check_rate()
        page = 1

        if not base_issues:
            milestone_number = None
            if milestone:
                milestone_data = self.milestone(milestone)
                milestone_number = milestone_data.get('number')
            issues = []
            while True:
                result = self.repo.issues.get(page=page,
                                              per_page=100,
                                              milestone=milestone_number,
                                              state=state,
                                              assignee=assignee,
                                              creator=creator,
                                              mentioned=mentioned,
                                              labels=labels,
                                              sort=sort,
                                              direction=direction,
                                              since=since)
                if len(result) > 0:
                    issues += result
                    page = page + 1
                else:
                    break
        else:
            issues = base_issues

        # If since was provided, filter the issue
        issues = self._filter_since(issues, since)

        # If until was provided, filter the issue
        issues = self._filter_until(issues, until)

        # If milestone was provided, filter the issue
        issues = self._filter_milestone(issues, milestone)

        # If it is a pr check if it is merged or closed, removed closed ones
        issues = self._filer_closed_prs(issues, branch)

        return issues

    def is_merged(self, pr):
        """
        Return wether a PR was merged, or if it was closed and discarded.

        https://developer.github.com/v3/pulls/#get-if-a-pull-request-has-been-merged
        """
        self._check_rate()
        merged = True
        try:
            self.repo('pulls')(str(pr))('merge').get()
        except Exception:
            merged = False
        return merged

    @staticmethod
    def str_to_date(string):
        """Convert ISO date string to datetime object."""
        parts = string.split('T')
        date_parts = parts[0]
        time_parts = parts[1][:-1]
        year, month, day = [int(i) for i in date_parts.split('-')]
        hour, minutes, seconds = [int(i) for i in time_parts.split(':')]
        return datetime.datetime(year, month, day, hour, minutes, seconds)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Loghub filter and formatter."""

# yapf: disable

# Standard library imports
from collections import OrderedDict
import codecs
import re
import sys
import time

# Third party imports
from jinja2 import Template

# Local imports
from loghub.core.repo import GitHubRepo
from loghub.core.zenhub import ZenHub
from loghub.templates import (CHANGELOG_GROUPS_TEMPLATE_PATH,
                              CHANGELOG_ISSUE_GROUPS_TEMPLATE_PATH,
                              CHANGELOG_PR_GROUPS_TEMPLATE_PATH,
                              CHANGELOG_TEMPLATE_PATH,
                              RELEASE_GROUPS_TEMPLATE_PATH,
                              RELEASE_ISSUE_GROUPS_TEMPLATE_PATH,
                              RELEASE_PR_GROUPS_TEMPLATE_PATH,
                              RELEASE_TEMPLATE_PATH)

# yapf: enable


def filter_issues_fixed_by_prs(issues, prs, show_related_prs, show_related_issues):
    """
    Find related issues to prs and prs to issues that are fixed.

    This adds extra information to the issues and prs listings.
    """
    words = [
        'close', 'closes', 'fix', 'fixes', 'fixed', 'resolve', 'resolves',
        'resolved'
    ]
    pattern = re.compile(
        r'(?P<word>' + r'|'.join(words) + r') '
        r'((?P<repo>.*?)#(?P<number>\d*)|(?P<full_repo>.*)/(?P<number_2>\d*))',
        re.IGNORECASE, )
    issue_pr_map = {}
    pr_issue_map = {}
    for pr in prs:
        is_pr = bool(pr.get('pull_request'))
        if is_pr:
            pr_url = pr.html_url
            pr_number = pr.number
            user = pr.user
            repo_url = pr_url.split('/pull/')[0] + '/issues/'
            pr_issue_map[pr_url] = []
            body = pr.body or ''

            # Remove blanks and markdown comments
            if body:
                lines = body.splitlines()
                no_comments = [l for l in lines
                               if (l and not l.startswith("<!---"))]
                body = '\n'.join(no_comments)

            for matches in pattern.finditer(body):
                dic = matches.groupdict()
                issue_number = dic['number'] or dic['number_2'] or ''
                repo = dic['full_repo'] or dic['repo'] or repo_url

                # Repo name can't have spaces.
                if ' ' not in repo:
                    # In case spyder-ide/loghub#45 was for example used
                    if 'http' not in repo:
                        repo = 'https://github.com/' + repo

                    if '/issues' not in repo:
                        issue_url = repo + '/issues/' + issue_number
                    elif repo.endswith('/') and issue_number:
                        issue_url = repo + issue_number
                    elif issue_number:
                        issue_url = repo + '/' + issue_number
                    else:
                        issue_url = None
                else:
                    issue_url = None

                # Set the issue data
                issue_data = {'url': pr_url, 'text': pr_number, 'user': user}
                if issue_url is not None:
                    if issue_number in issue_pr_map:
                        issue_pr_map[issue_url].append(issue_data)
                    else:
                        issue_pr_map[issue_url] = [issue_data]

                    pr_data = {'url': issue_url, 'text': issue_number}
                    pr_issue_map[pr_url].append(pr_data)

            if show_related_issues:
                pr['loghub_related_issues'] = pr_issue_map[pr_url]

    for issue in issues:
        issue_url = issue.html_url
        if issue_url in issue_pr_map and show_related_prs:
            issue['loghub_related_pulls'] = issue_pr_map[issue_url]

    # Now sort the numbers in descending order
    for issue in issues:
        related_pulls = issue.get('loghub_related_pulls', [])
        related_pulls = sorted(
            related_pulls, key=lambda p: p['url'], reverse=True)
        issue['loghub_related_pulls'] = related_pulls

    for pr in prs:
        related_issues = pr.get('loghub_related_issues', [])
        related_issues = sorted(
            related_issues, key=lambda i: i['url'], reverse=True)
        pr['loghub_related_issues'] = related_issues

    return issues, prs


def filter_prs_by_regex(issues, pr_label_regex):
    """Filter prs by issue regex."""
    filtered_prs = []
    pr_pattern = re.compile(pr_label_regex)

    for issue in issues:
        is_pr = bool(issue.get('pull_request'))
        labels = ' '.join(issue.get('loghub_label_names'))

        if is_pr:
            if pr_label_regex:
                pr_valid = bool(pr_pattern.search(labels))
                if pr_valid:
                    filtered_prs.append(issue)
            else:
                filtered_prs.append(issue)

    return filtered_prs


def filter_issues_by_regex(issues, issue_label_regex):
    """Filter issues by issue regex."""
    filtered_issues = []
    issue_pattern = re.compile(issue_label_regex)

    for issue in issues:
        is_pr = bool(issue.get('pull_request'))
        is_issue = not is_pr
        labels = ' '.join(issue.get('loghub_label_names'))

        if is_issue and issue_label_regex:
            issue_valid = bool(issue_pattern.search(labels))
            if issue_valid:
                filtered_issues.append(issue)
        elif is_issue and not issue_label_regex:
            filtered_issues.append(issue)

    return filtered_issues


def filter_issue_label_groups(issues, issue_label_groups):
    """Filter issues by the label groups."""
    grouped_filtered_issues = OrderedDict()
    if issue_label_groups:
        new_filtered_issues = []
        for label_group_dic in issue_label_groups:
            grouped_filtered_issues[label_group_dic['name']] = []

        for issue in issues:
            labels = issue.get('loghub_label_names')
            for label_group_dic in issue_label_groups:
                label = label_group_dic['label']
                name = label_group_dic['name']
                if label in labels:
                    grouped_filtered_issues[name].append(issue)
                    new_filtered_issues.append(issue)
    else:
        new_filtered_issues = issues

    return new_filtered_issues, grouped_filtered_issues


def join_label_groups(grouped_issues, grouped_prs, issue_label_groups,
                      pr_label_groups):
    """Combine issue and PR groups in to one dictionary.

    PR-only groups are added after all issue groups. Any groups that are
    shared between issues and PRs are added according to the order in the
    issues list of groups. This results in "label-groups" remaining in the
    same order originally specified even if a group does not have issues
    in it. Otherwise, a shared group may end up at the end of the combined
    dictionary and not in the order originally specified by the user.

    """
    issue_group_names = [x['name'] for x in issue_label_groups]
    pr_group_names = [x['name'] for x in pr_label_groups]
    shared_groups = []
    for idx, group_name in enumerate(issue_group_names):
        if len(pr_group_names) > idx and group_name == pr_group_names[idx]:
            shared_groups.append(group_name)
        else:
            break

    label_groups = OrderedDict()
    # add shared groups first
    for group_name in shared_groups:
        # make sure to copy the issue group in case it is added to
        label_groups[group_name] = grouped_issues.get(group_name, [])[:]
    # add any remaining issue groups
    for group_name, group in grouped_issues.items():
        if group_name in shared_groups:
            continue
        label_groups[group_name] = group[:]
    # add any remaining PR groups (extending any existing groups)
    for group_name, group in grouped_prs.items():
        label_groups.setdefault(group_name, []).extend(group)
    return label_groups


def create_changelog(repo=None,
                     username=None,
                     password=None,
                     token=None,
                     milestone=None,
                     zenhub_release=None,
                     zenhub_token=None,
                     since_tag=None,
                     until_tag=None,
                     branch=None,
                     output_format='changelog',
                     issue_label_regex='',
                     pr_label_regex='',
                     template_file=None,
                     issue_label_groups=None,
                     pr_label_groups=None,
                     batch=None,
                     show_prs=True,
                     show_related_prs=True,
                     show_related_issues=True):
    """Create changelog data for single and batched mode."""
    if issue_label_groups is None:
        issue_label_groups = []

    if pr_label_groups is None:
        pr_label_groups = []

    gh = GitHubRepo(
        username=username,
        password=password,
        token=token,
        repo=repo, )

    all_changelogs = []
    version_tag_prefix = 'v'

    base_issues = None
    if zenhub_release:
        items = [(None, None, None)]
    elif batch:
        # This will get all the issues, might eat up the api rate limit!
        base_issues = issues = gh.issues(state='closed', branch=branch)
        if batch == 'milestones':
            milestones = [i.get('title') for i in gh.milestones()]
            empty_items = [None] * len(milestones)
            items = list(zip(milestones, empty_items, empty_items))
        elif batch == 'tags':
            tags = [
                i.get('ref', '').replace('refs/tags/', '') for i in gh.tags()
            ]
            since_tags = [None] + tags
            until_tags = tags + [None]
            empty_items = [None] * len(since_tags)
            items = list(zip(empty_items, since_tags, until_tags))
    else:
        if milestone:
            items = [(milestone, None, None)]
        else:
            items = [(None, since_tag, until_tag)]

    for (milestone, since_tag, until_tag) in reversed(items):
        version = until_tag or None
        closed_at = None
        since = None
        until = None

        # Set milestone or from tag
        if milestone and not since_tag:
            milestone_data = gh.milestone(milestone)
            closed_at = milestone_data['closed_at']
            version = milestone

            if version.startswith(version_tag_prefix):
                version = version[len(version_tag_prefix):]

        elif not milestone and since_tag:
            since = gh.tag(since_tag)['tagger']['date']
            if until_tag:
                until = gh.tag(until_tag)['tagger']['date']
                closed_at = until

        if not bool(zenhub_release):
            # This returns issues and pull requests
            issues = gh.issues(
                milestone=milestone,
                state='closed',
                since=since,
                until=until,
                branch=branch,
                base_issues=base_issues,
            )
        else:
            version = zenhub_release
            zh = ZenHub(zenhub_token)

            # Get repo id
            repo_id = gh.repo.get()['id']

            # Get list of releases and select the right one
            releases = zh.releases(repo_id)
            release_id = None
            for release in releases:
                if zenhub_release == release['title']:
                    release_id = release['release_id']
                    break

            # Get all the specific issues 1 by 1
            if release_id is not None:
                zh_issues = zh.issues(release_id)

                # Filter issues that belong to this repository
                zh_issues = [issue for issue in zh_issues if issue['repo_id'] == repo_id]
                
                # Get all issues from github
                issues = []
                print('\n{} issues found!'.format(len(zh_issues)))
                for idx, zh_issue in enumerate(reversed(zh_issues)):
                    print(idx)
                    issue = gh.issue(zh_issue['issue_number'])
                    if issue['state'] == 'closed':
                        # Add filtered label names inside additional key
                        issue['loghub_label_names'] = [l['name'] for l in issue.get('labels')]
                        issues.append(issue)
            else:
                release_titles = [release['title'] for release in releases]
                print("Zenhub release not found! Available releases are: {}".format(
                    release_titles))
                sys.exit(1)

        # Filter by regex if available
        filtered_prs = filter_prs_by_regex(issues, pr_label_regex)
        filtered_issues = filter_issues_by_regex(issues, issue_label_regex)

        # If issue label grouping, filter issues
        filtered_issues, grouped_issues = filter_issue_label_groups(
            filtered_issues, issue_label_groups)
        filtered_prs, grouped_prs = filter_issue_label_groups(
            filtered_prs, pr_label_groups)
        label_groups = join_label_groups(grouped_issues, grouped_prs,
                                         issue_label_groups, pr_label_groups)

        filter_issues_fixed_by_prs(filtered_issues, filtered_prs,
                                   show_related_prs, show_related_issues)

        ch = render_changelog(
            repo,
            filtered_issues,
            filtered_prs,
            version,
            closed_at=closed_at,
            output_format=output_format,
            template_file=template_file,
            label_groups=label_groups,
            issue_label_groups=grouped_issues,
            pr_label_groups=grouped_prs,
            show_prs=show_prs)

        all_changelogs.append(ch)

    changelog = '\n'.join(all_changelogs)
    write_changelog(changelog=changelog)

    return changelog


def render_changelog(repo,
                     issues,
                     prs,
                     version=None,
                     closed_at=None,
                     output_format='changelog',
                     template_file=None,
                     issue_label_groups=None,
                     pr_label_groups=None,
                     label_groups=None,
                     show_prs=True):
    """Render changelog data on a jinja template."""
    # Header
    if not version:
        version = '<RELEASE_VERSION>'

    if closed_at:
        close_date = closed_at.split('T')[0]
    else:
        close_date = time.strftime("%Y/%m/%d")

    # Load template
    if template_file:
        filepath = template_file
    else:
        if issue_label_groups and pr_label_groups:
            if output_format == 'changelog':
                filepath = CHANGELOG_GROUPS_TEMPLATE_PATH
            else:
                filepath = RELEASE_GROUPS_TEMPLATE_PATH
        elif issue_label_groups:
            if output_format == 'changelog':
                filepath = CHANGELOG_ISSUE_GROUPS_TEMPLATE_PATH
            else:
                filepath = RELEASE_ISSUE_GROUPS_TEMPLATE_PATH
        elif pr_label_groups:
            if output_format == 'changelog':
                filepath = CHANGELOG_PR_GROUPS_TEMPLATE_PATH
            else:
                filepath = RELEASE_PR_GROUPS_TEMPLATE_PATH
        else:
            if output_format == 'changelog':
                filepath = CHANGELOG_TEMPLATE_PATH
            else:
                filepath = RELEASE_TEMPLATE_PATH

    with open(filepath) as f:
        data = f.read()

    repo_owner, repo_name = repo.split('/')
    template = Template(data)
    rendered = template.render(
        issues=issues,
        pull_requests=prs,
        version=version,
        close_date=close_date,
        repo_full_name=repo,
        repo_owner=repo_owner,
        repo_name=repo_name,
        label_groups=label_groups,
        issue_label_groups=issue_label_groups,
        pr_label_groups=pr_label_groups,
        show_prs=show_prs)

    return rendered


def write_changelog(changelog, output_file='CHANGELOG.temp'):
    """Output rendered result to prompt and file."""
    print('#' * 79)
    print(changelog)
    print('#' * 79)

    with codecs.open(output_file, "w", "utf-8") as f:
        f.write(changelog)
